package scanner;

import static util.Characters.BACKSLASH;
import static util.Characters.DOUBLE_QUOTE;
import static util.Characters.EOF;
import static util.Characters.NEWLINE;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;

import constructs.Token;
import constructs.TokenFactory;
import constructs.TokenFactoryImpl;
import error.LispException;
import file.FilePosition;
import file.FilePositionTracker;
import util.Characters;

/**
 * Converts a stream of bytes into a stream of Lisp tokens.
 */
public class LispScanner {

    private InputStream inputStream;
    private FilePositionTracker positionTracker;
    private TokenFactory tokenFactory;

    public LispScanner(InputStream in, String fileName) {
        this.inputStream = new LispFilterInputStream(new BufferedInputStream(in));
        this.positionTracker = new FilePositionTracker(fileName);
        this.tokenFactory = new TokenFactoryImpl();
    }

    public Token getNextToken() throws IOException {
        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;
            positionTracker.incrementColumn();

            if (Character.isWhitespace(nextChar)) {
                if (nextChar == NEWLINE)
                    positionTracker.incrementLine();
            } else {
                FilePosition currentPosition = positionTracker.getCurrentPosition();
                String tokenText = retrieveTokenText(nextChar);

                return tokenFactory.createToken(tokenText, currentPosition);
            }
        }

        return tokenFactory.createEOFToken(positionTracker.getCurrentPosition());
    }

    private String retrieveTokenText(char firstCharacter) throws IOException {
        String tokenText = "" + firstCharacter;

        if (firstCharacter == DOUBLE_QUOTE)
            tokenText = retrieveString(firstCharacter);
        else if (Character.isDigit(firstCharacter))
            tokenText = retrieveNumber(firstCharacter);
        else if (Characters.isLegalIdentifierCharacter(firstCharacter))
            tokenText = retrieveIdentifier(firstCharacter);

        return tokenText;
    }

    private String retrieveString(char firstDoubleQuote) throws IOException {
        StringBuilder text = new StringBuilder();
        FilePosition stringPosition = positionTracker.getCurrentPosition();
        char prevChar = firstDoubleQuote;

        text.append(firstDoubleQuote);

        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            positionTracker.incrementColumn();
            text.append(nextChar);

            switch (nextChar) {
            case NEWLINE:
                positionTracker.incrementLine();
                break;
            case DOUBLE_QUOTE:
                if (prevChar != BACKSLASH) {
                    // we have found the terminating double quote

                    return text.toString();
                }

                // this is an escaped double quote
            }

            prevChar = nextChar;
        }

        throw new UnterminatedStringException(stringPosition);
    }

    private String retrieveNumber(char firstDigit) throws IOException {
        StringBuilder text = new StringBuilder();

        text.append(firstDigit);
        inputStream.mark(1);

        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            if (Character.isDigit(nextChar)) {
                // 'nextChar' is a digit in this number

                text.append(nextChar);
                positionTracker.incrementColumn();
            } else {
                // we have reached the end of the number

                inputStream.reset(); // unread the last character

                return text.toString();
            }

            inputStream.mark(1);
        }

        return text.toString();
    }

    private String retrieveIdentifier(char firstChar) throws IOException {
        StringBuilder text = new StringBuilder();

        text.append(firstChar);
        inputStream.mark(1);

        int c;

        while ((c = inputStream.read()) != EOF) {
            char nextChar = (char) c;

            if (Characters.isLegalIdentifierCharacter(nextChar)) {
                // 'nextChar' is part of the identifier

                text.append(nextChar);
                positionTracker.incrementColumn();
            } else {
                // we have reached the end of this identifier

                inputStream.reset(); // unread the last character

                return text.toString();
            }

            inputStream.mark(1);
        }

        return text.toString();
    }

    public static class UnterminatedStringException extends LispException {

        private static final long serialVersionUID = 1L;
        private FilePosition position;

        public UnterminatedStringException(FilePosition position) {
            this.position = position;
        }

        @Override
        public int getSeverity() {
            return 0;
        }

        @Override
        public String getMessage() {
            return MessageFormat.format("unterminated quoted string - line {0}, column {1}", position.getLineNumber(),
                                        position.getColumnNumber());
        }

    }

}
